"""
The latest version of this package is available at:
<http://github.com/jantman/biweeklybudget>

################################################################################
Copyright 2016-2024 Jason Antman <http://www.jasonantman.com>

    This file is part of biweeklybudget, also known as biweeklybudget.

    biweeklybudget is free software: you can redistribute it and/or modify
    it under the terms of the GNU Affero General Public License as published by
    the Free Software Foundation, either version 3 of the License, or
    (at your option) any later version.

    biweeklybudget is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU Affero General Public License for more details.

    You should have received a copy of the GNU Affero General Public License
    along with biweeklybudget.  If not, see <http://www.gnu.org/licenses/>.

The Copyright and Authors attributions contained herein may not be removed or
otherwise altered, except to add the Author attribution of a contributor to
this work. (Additional Terms pursuant to Section 7b of the AGPL v3)
################################################################################
While not legally required, I sincerely request that anyone who finds
bugs please submit them at <https://github.com/jantman/biweeklybudget> or
to me via email, and that you send any contributions or improvements
either as a pull request on GitHub, or to me via email.
################################################################################

AUTHORS:
Jason Antman <jason@jasonantman.com> <http://www.jasonantman.com>
################################################################################
"""

import os
import logging
from textwrap import dedent
from flask.views import MethodView
from flask import jsonify, request, redirect, render_template
from plaid import ApiException
from typing import List, Dict

from biweeklybudget import settings
from biweeklybudget.flaskapp.app import app
from biweeklybudget.utils import plaid_client, dtnow
from biweeklybudget.models.plaid_items import PlaidItem
from biweeklybudget.models.plaid_accounts import PlaidAccount
from biweeklybudget.plaid_updater import PlaidUpdater
from biweeklybudget.version import VERSION
from biweeklybudget.db import db_session

from plaid.models import (
    LinkTokenCreateRequest, ItemPublicTokenExchangeRequest,
    LinkTokenCreateRequestUser, ItemGetRequest, InstitutionsGetByIdRequest,
    AccountsGetRequest, LinkTokenCreateResponse, Products, CountryCode
)

logger = logging.getLogger(__name__)


class PlaidJs(MethodView):
    """
    Handle GET /plaid.js endpoint, for CI/test or production/real.
    """

    def get(self):
        return redirect('static/js/plaid_prod.js')


class PlaidHandleLink(MethodView):
    """
    Handle POST /ajax/plaid/handle_link endpoint.
    """

    def post(self):
        data = request.get_json()
        logger.debug(
            'got POST to /ajax/plaid/handle_link; data=%s', data
        )
        client = plaid_client()
        public_token = data['public_token']
        logger.debug('Plaid token exchange for public token: %s', public_token)
        try:
            req = ItemPublicTokenExchangeRequest(public_token=public_token)
            exchange_response = client.item_public_token_exchange(req)
        except ApiException as e:
            logger.error(
                'Plaid error exchanging token %s: %s',
                public_token, e, exc_info=True
            )
            resp = jsonify({
                'success': False,
                'message': 'Exception: %s' % str(e)
            })
            resp.status_code = 400
            return resp
        logger.info(
            'Plaid token exchange: public_token=%s response=%s',
            public_token, exchange_response
        )
        item = PlaidItem(
            item_id=exchange_response['item_id'],
            access_token=exchange_response['access_token'],
            last_updated=dtnow(),
            institution_name=data['metadata']['institution']['name'],
            institution_id=data['metadata']['institution']['institution_id']
        )
        logger.info('Add to DB: %s', item)
        db_session.add(item)
        for acct in data['metadata']['accounts']:
            acct = PlaidAccount(
                item_id=exchange_response['item_id'],
                account_id=acct['id'],
                name=acct['name'],
                mask=acct['mask'],
                account_type=acct['type'],
                account_subtype=acct.get('subtype', 'unknown')
            )
            logger.info('Add to DB: %s', acct)
            db_session.add(acct)
        db_session.commit()
        return jsonify({
            'success': True,
            'item_id': exchange_response['item_id'],
            'access_token': exchange_response['access_token']
        })


class PlaidRefreshAccounts(MethodView):
    """
    Handle POST /ajax/plaid/refresh_item_accounts endpoint.
    """

    def post(self):
        data = request.get_json()
        client = plaid_client()
        item_id = data['item_id']
        item: PlaidItem = db_session.query(PlaidItem).get(item_id)
        logger.debug(
            'Plaid refresh accounts for %s', item
        )
        try:
            response = client.accounts_get(
                AccountsGetRequest(access_token=item.access_token)
            )
        except ApiException as e:
            logger.error(
                'Plaid error getting accounts for %s: %s',
                item.access_token, e, exc_info=True
            )
            resp = jsonify({
                'success': False,
                'message': 'Exception: %s' % str(e)
            })
            resp.status_code = 400
            return resp
        logger.info('Plaid account refresh for item %s: %s', item, response)
        a: PlaidAccount
        all_accts = {a.account_id: a for a in item.all_accounts}
        curr_acct_ids = []
        logger.info(
            'Plaid reported %d accounts for the item', len(response['accounts'])
        )
        for acct in response['accounts']:
            acct = acct.to_dict()
            if acct['account_id'] in all_accts:
                curr_acct_ids.append(acct['account_id'])
                continue
            a = PlaidAccount(
                item_id=item.item_id,
                account_id=acct['account_id'],
                name=acct['name'],
                mask=acct['mask'],
                account_type=acct['type'],
                account_subtype=acct.get('subtype', 'unknown')
            )
            curr_acct_ids.append(acct['account_id'])
            logger.info('Add new to DB: %s', a)
            db_session.add(a)
        for aid, a in all_accts.items():
            if aid in curr_acct_ids:
                continue
            logger.info('Account no longer in Plaid item; removing: %s', a)
            db_session.delete(a)
        db_session.commit()
        return jsonify({'success': True})


class PlaidUpdateItemInfo(MethodView):
    """
    Handle POST /ajax/plaid/update_item_info endpoint.
    """

    def post(self):
        client = plaid_client()
        logger.info('Refreshing Plaid item info')
        item: PlaidItem
        for item in db_session.query(PlaidItem).all():
            logger.debug(
                'Plaid refresh item: %s', item
            )
            try:
                response = client.item_get(
                    ItemGetRequest(access_token=item.access_token)
                )
            except ApiException as e:
                logger.error(
                    'Plaid error getting item %s: %s',
                    item, e, exc_info=True
                )
                resp = jsonify({
                    'success': False,
                    'message': 'Exception: %s' % str(e)
                })
                resp.status_code = 400
                return resp
            logger.info('Plaid item info item %s: %s', item, response)
            item.institution_id = response['item']['institution_id']
            inst = client.institutions_get_by_id(
                InstitutionsGetByIdRequest(
                    institution_id=response['item']['institution_id'],
                    country_codes=[
                        CountryCode(x)
                        for x in settings.PLAID_COUNTRY_CODES.split(',')
                    ]
                )
            )
            item.institution_name = inst['institution']['name']
            db_session.add(item)
        db_session.commit()
        return jsonify({'success': True})


class PlaidUpdate(MethodView):
    """
    Handle GET or POST /plaid-update

    This single endpoint has multiple functions:

    * If GET with no query parameters, displays a form template to use to
      interactively update Plaid accounts.
    * If GET or POST with an ``item_ids`` query parameter, performs a Plaid
      update (via :py:meth:`~._update`) of the specified CSV list of Plaid Item
      IDs, or all Plaid Items if the value is ``ALL``. The POST method also
      accepts an optional ``num_days`` parameter specifying an integer number of
      days of transactions to update. The response from this endpoint can be in
      one of three forms:

      * If the ``Accept`` HTTP header is set to ``application/json``, return a
        JSON list of update results. Each list item is the JSON-ified value of
        :py:attr:`~.PlaidUpdateResult.as_dict`.
      * If the ``Accept`` HTTP header is set to ``text/plain``, return a plain
        text human-readable summary of the update operation.
      * Otherwise, return a templated view of the update operation results, as
        would be returned to a browser.
    """

    def post(self):
        """
        Handle POST. If the ``item_ids`` query parameter is set, then return
        :py:meth:`~._update`, else return a HTTP 400. If the optional
        ``num_days`` query parameter is set, pass that on to the update method.
        """
        ids = request.args.get('item_ids')
        if ids is None and request.form:
            ids = ','.join([
                x.replace('item_', '') for x in request.form.keys()
            ])
        if ids is None:
            return jsonify({
                'success': False,
                'message': 'Missing parameter: item_ids'
            }), 400
        kwargs = {}
        if 'num_days' in request.args:
            kwargs['num_days'] = int(request.args['num_days'])
        return self._update(ids, **kwargs)

    def get(self):
        """
        Handle GET. If the ``item_ids`` query parameter is set, then return
        :py:meth:`~._update`, else return :py:meth:`~._form`.
        """
        ids = request.args.get('item_ids')
        if ids is None:
            return self._form()
        kwargs = {}
        if 'num_days' in request.args:
            kwargs['num_days'] = int(request.args['num_days'])
        return self._update(ids, **kwargs)

    def _update(self, ids: str, num_days: int = 30):
        """
        Handle an update for Plaid accounts by instantiating a
        :py:class:`~.PlaidUpdater`, calling its :py:meth:`~.PlaidUpdater.update`
        method with the proper arguments, and then returning the result in a
        form determined by the ``Accept`` header.

        :param ids: a comma-separated string listing the :py:class:`~.PlaidItem`
          IDs to update, or the string ``ALL`` to update all Items.
        :type ids: str
        :param num_days: number of days to retrieve transactions for; default 30
        :type num_days: int
        """
        logger.info(
            'Handle Plaid Update request; item_ids=%s num_days=%d',
            ids, num_days
        )
        updater = PlaidUpdater()
        if ids == 'ALL':
            items = PlaidUpdater.available_items()
        else:
            ids = ids.split(',')
            items = [
                db_session.query(PlaidItem).get(x) for x in ids
            ]
        results = updater.update(items=items, days=num_days)
        if request.headers.get('accept') == 'text/plain':
            s = ''
            num_updated = 0
            num_added = 0
            num_failed = 0
            for r in results:
                if not r.success:
                    num_failed += 1
                    s += f'{r.item.institution_name} ({r.item.item_id}): ' \
                         f'Failed: {r.exc}\n'
                    continue
                num_updated += r.updated
                num_added += r.added
                s += f'{r.item.institution_name} ({r.item.item_id}): ' \
                     f'{r.updated} ' \
                     f'updated, {r.added} added (stmts: {r.stmt_ids})\n'
            s += f'TOTAL: {num_updated} updated, {num_added} added, ' \
                 f'{num_failed} account(s) failed'
            return s
        if request.headers.get('accept') == 'application/json':
            return jsonify([x.as_dict for x in results])
        # have to do this here in python and iterate twice, because of
        # https://github.com/pallets/jinja/issues/641
        num_updated = 0
        num_added = 0
        num_failed = 0
        for r in results:
            num_updated += r.updated
            num_added += r.added
            if not r.success:
                num_failed += 1
        return render_template(
            'plaid_result.html',
            results=results,
            num_added=num_added,
            num_updated=num_updated,
            num_failed=num_failed
        )

    def _form(self):
        items: List[PlaidItem] = db_session.query(PlaidItem).all()
        x: PlaidItem
        a: PlaidAccount
        plaid_accounts: Dict[str, str] = {
            x.item_id: ', '.join(
                [
                    f'{a.name} ({a.mask})'
                    for a in sorted(x.all_accounts, key=lambda a: a.name)
                ]
            ) for x in items
        }
        accounts = {
            x.item_id: ', '.join(
                filter(
                    None,
                    [
                        None if a.account is None else
                        f'{a.account.name} ({a.account.id})'
                        for a in x.all_accounts
                    ]
                )
            ) for x in items
        }
        return render_template(
            'plaid_form.html', plaid_items=items, plaid_accounts=plaid_accounts,
            accounts=accounts
        )


class PlaidLinkToken(MethodView):
    """
    Handle POST /ajax/plaid/create_link_token endpoint.
    """

    def post(self):
        data = request.get_json()
        client = plaid_client()
        logger.debug('Plaid create link token')
        kwargs = dict(
            products=[
                Products(x) for x in settings.PLAID_PRODUCTS.split(',')
            ],
            client_name=f'github.com/jantman/biweeklybudget {VERSION}',
            country_codes=[
                CountryCode(x)
                for x in settings.PLAID_COUNTRY_CODES.split(',')
            ],
            language="en",
            user=LinkTokenCreateRequestUser(
                client_user_id=settings.PLAID_USER_ID
            ),
        )
        if 'item_id' in data:
            item_id = data['item_id']
            item: PlaidItem = db_session.query(PlaidItem).get(item_id)
            logger.info('PlaidLinkToken for updating item %s', item_id)
            kwargs['access_token'] = item.access_token
        try:
            req = LinkTokenCreateRequest(**kwargs)
            response: LinkTokenCreateResponse = client.link_token_create(req)
            logger.debug('Plaid link_token_create response: %s', response)
        except ApiException as e:
            logger.error(
                'Plaid error creating Link token: %s', e, exc_info=True
            )
            resp = jsonify({
                'success': False,
                'message': 'Exception: %s' % str(e)
            })
            resp.status_code = 400
            return resp
        logger.info(
            'Plaid Link token creation: %s', response
        )
        return jsonify({'link_token': response.link_token})


def set_url_rules(a):
    a.add_url_rule(
        '/ajax/plaid/handle_link',
        view_func=PlaidHandleLink.as_view('plaid_handle_link')
    )
    a.add_url_rule('/plaid.js', view_func=PlaidJs.as_view('plaid_js'))
    a.add_url_rule(
        '/plaid-update',
        view_func=PlaidUpdate.as_view('plaid_update')
    )
    a.add_url_rule(
        '/ajax/plaid/refresh_item_accounts',
        view_func=PlaidRefreshAccounts.as_view('plaid_refresh_item_accounts')
    )
    a.add_url_rule(
        '/ajax/plaid/update_item_info',
        view_func=PlaidUpdateItemInfo.as_view('plaid_update_item_info')
    )
    a.add_url_rule(
        '/ajax/plaid/create_link_token',
        view_func=PlaidLinkToken.as_view('plaid_link_token')
    )


set_url_rules(app)



"""
The latest version of this package is available at:
<http://github.com/jantman/biweeklybudget>

################################################################################
Copyright 2016-2024 Jason Antman <http://www.jasonantman.com>

    This file is part of biweeklybudget, also known as biweeklybudget.

    biweeklybudget is free software: you can redistribute it and/or modify
    it under the terms of the GNU Affero General Public License as published by
    the Free Software Foundation, either version 3 of the License, or
    (at your option) any later version.

    biweeklybudget is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU Affero General Public License for more details.

    You should have received a copy of the GNU Affero General Public License
    along with biweeklybudget.  If not, see <http://www.gnu.org/licenses/>.

The Copyright and Authors attributions contained herein may not be removed or
otherwise altered, except to add the Author attribution of a contributor to
this work. (Additional Terms pursuant to Section 7b of the AGPL v3)
################################################################################
While not legally required, I sincerely request that anyone who finds
bugs please submit them at <https://github.com/jantman/biweeklybudget> or
to me via email, and that you send any contributions or improvements
either as a pull request on GitHub, or to me via email.
################################################################################

AUTHORS:
Jason Antman <jason@jasonantman.com> <http://www.jasonantman.com>
################################################################################
"""

from sqlalchemy import Column, Integer, Numeric, ForeignKey
from sqlalchemy.orm import relationship
from biweeklybudget.models.base import Base, ModelAsDict


class BudgetTransaction(Base, ModelAsDict):
    """
    Represents the portion (amount) of a Transaction that is allocated
    against a specific budget. There will be one or more BudgetTransactions
    associated with each :py:class:`~.Transaction`.
    """

    __tablename__ = 'budget_transactions'
    __table_args__ = (
        {'mysql_engine': 'InnoDB'}
    )

    #: Primary Key
    id = Column(Integer, primary_key=True)

    #: Amount of the transaction against this budget
    amount = Column(Numeric(precision=10, scale=4), nullable=False)

    #: ID of the Transaction this is part of
    trans_id = Column(Integer, ForeignKey('transactions.id'))

    #: Relationship - the :py:class:`~.Transaction` this is part of
    transaction = relationship(
        "Transaction", backref="budget_transactions", uselist=False
    )

    #: ID of the Budget this transaction is against
    budget_id = Column(Integer, ForeignKey('budgets.id'))

    #: Relationship - the :py:class:`~.Budget` this transaction is against
    budget = relationship(
        "Budget", backref="budget_transactions", uselist=False
    )

    def __repr__(self):
        return "<BudgetTransaction(id=%s, transaction=%s, " \
               "budget=%s, amount=%s)>" % (
                   self.id, self.transaction, self.budget, self.amount
               )



"""add weekly and annual scheduled transaction types

Revision ID: bbb39e1d7c5d
Revises: 1bb9e6a1c07c
Create Date: 2026-01-19

"""
from alembic import op
import sqlalchemy as sa


# revision identifiers, used by Alembic.
revision = 'bbb39e1d7c5d'
down_revision = '1bb9e6a1c07c'
branch_labels = None
depends_on = None


def upgrade():
    # Add day_of_week column for weekly scheduled transactions (0=Monday, 6=Sunday)
    op.add_column(
        'scheduled_transactions',
        sa.Column('day_of_week', sa.SmallInteger(), nullable=True)
    )
    # Add annual_month column for annual scheduled transactions (1-12)
    op.add_column(
        'scheduled_transactions',
        sa.Column('annual_month', sa.SmallInteger(), nullable=True)
    )
    # Add annual_day column for annual scheduled transactions (1-31)
    op.add_column(
        'scheduled_transactions',
        sa.Column('annual_day', sa.SmallInteger(), nullable=True)
    )


def downgrade():
    op.drop_column('scheduled_transactions', 'annual_day')
    op.drop_column('scheduled_transactions', 'annual_month')
    op.drop_column('scheduled_transactions', 'day_of_week')



"""
The latest version of this package is available at:
<http://github.com/jantman/biweeklybudget>

################################################################################
Copyright 2016-2024 Jason Antman <http://www.jasonantman.com>

    This file is part of biweeklybudget, also known as biweeklybudget.

    biweeklybudget is free software: you can redistribute it and/or modify
    it under the terms of the GNU Affero General Public License as published by
    the Free Software Foundation, either version 3 of the License, or
    (at your option) any later version.

    biweeklybudget is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU Affero General Public License for more details.

    You should have received a copy of the GNU Affero General Public License
    along with biweeklybudget.  If not, see <http://www.gnu.org/licenses/>.

The Copyright and Authors attributions contained herein may not be removed or
otherwise altered, except to add the Author attribution of a contributor to
this work. (Additional Terms pursuant to Section 7b of the AGPL v3)
################################################################################
While not legally required, I sincerely request that anyone who finds
bugs please submit them at <https://github.com/jantman/biweeklybudget> or
to me via email, and that you send any contributions or improvements
either as a pull request on GitHub, or to me via email.
################################################################################

AUTHORS:
Jason Antman <jason@jasonantman.com> <http://www.jasonantman.com>
################################################################################
"""

import logging
from flask.views import MethodView
from flask import render_template, jsonify
from decimal import Decimal
from datetime import datetime
import json
import re

from biweeklybudget.flaskapp.app import app
from biweeklybudget.flaskapp.views.formhandlerview import FormHandlerView
from biweeklybudget.models.account import Account, AcctType
from biweeklybudget.models.budget_model import Budget
from biweeklybudget.models.plaid_accounts import PlaidAccount
from biweeklybudget.models.plaid_items import PlaidItem
from biweeklybudget.models.transaction import Transaction
from biweeklybudget.db import db_session
from biweeklybudget.interest import (
    INTEREST_CALCULATION_NAMES, MIN_PAYMENT_FORMULA_NAMES
)

logger = logging.getLogger(__name__)

RE_FIELD_NAMES = [
    're_interest_charge',
    're_interest_paid',
    're_payment',
    're_late_fee',
    're_other_fee'
]


class AccountsView(MethodView):
    """
    Render the GET /accounts view using the ``accounts.html`` template.
    """

    def get(self):
        accts = {a.name: a.id for a in db_session.query(Account).all()}
        budgets = {}
        active_budgets = {}
        for b in db_session.query(Budget).all():
            k = b.name
            if b.is_income:
                k = '%s (i)' % b.name
            budgets[b.id] = k
            if b.is_active:
                active_budgets[b.id] = k
        pa: PlaidAccount
        plaid_accts = {
            f'{pa.plaid_item.institution_name} / {pa.name} ({pa.mask})':
                f'{pa.item_id},{pa.account_id}'
            for pa in db_session.query(PlaidAccount).all()
        }
        return render_template(
            'accounts.html',
            bank_accounts=db_session.query(Account).filter(
                Account.acct_type == AcctType.Bank,
                Account.is_active == True).all(),  # noqa
            credit_accounts=db_session.query(Account).filter(
                Account.acct_type == AcctType.Credit,
                Account.is_active == True).all(),  # noqa
            investment_accounts=db_session.query(Account).filter(
                Account.acct_type == AcctType.Investment,
                Account.is_active == True).all(),  # noqa
            interest_class_names=INTEREST_CALCULATION_NAMES.keys(),
            min_pay_class_names=MIN_PAYMENT_FORMULA_NAMES.keys(),
            accts=accts,
            budgets=budgets,
            active_budgets=active_budgets,
            plaid_accounts=plaid_accts
        )


class OneAccountView(MethodView):
    """
    Render the /accounts/<int:acct_id> view using the ``account.html``
    template.
    """

    def get(self, acct_id):
        accts = {a.name: a.id for a in db_session.query(Account).all()}
        budgets = {}
        active_budgets = {}
        for b in db_session.query(Budget).all():
            k = b.name
            if b.is_income:
                k = '%s (i)' % b.name
            budgets[b.id] = k
            if b.is_active:
                active_budgets[b.id] = k
        pa: PlaidAccount
        plaid_accts = {
            f'{pa.plaid_item.institution_name} / {pa.name} ({pa.mask})':
                f'{pa.item_id},{pa.account_id}'
            for pa in db_session.query(PlaidAccount).all()
        }
        return render_template(
            'accounts.html',
            bank_accounts=db_session.query(Account).filter(
                Account.acct_type == AcctType.Bank,
                Account.is_active == True).all(),  # noqa
            credit_accounts=db_session.query(Account).filter(
                Account.acct_type == AcctType.Credit,
                Account.is_active == True).all(),  # noqa
            investment_accounts=db_session.query(Account).filter(
                Account.acct_type == AcctType.Investment,
                Account.is_active == True).all(),  # noqa
            account_id=acct_id,
            interest_class_names=INTEREST_CALCULATION_NAMES.keys(),
            min_pay_class_names=MIN_PAYMENT_FORMULA_NAMES.keys(),
            accts=accts,
            budgets=budgets,
            active_budgets=active_budgets,
            plaid_accounts=plaid_accts
        )


class AccountAjax(MethodView):
    """
    Handle GET /ajax/account/<int:account_id> endpoint.
    """

    def get(self, account_id):
        acct = db_session.query(Account).get(account_id)
        return jsonify(acct.as_dict)


class AccountFormHandler(FormHandlerView):
    """
    Handle POST /forms/account
    """

    def validate(self, data):
        """
        Validate the form data. Return None if it is valid, or else a hash of
        field names to list of error strings for each field.

        :param data: submitted form data
        :type data: dict
        :return: None if no errors, or hash of field name to errors for that
          field
        """
        have_errors = False
        errors = {k: [] for k in data.keys()}

        if data.get('name', '').strip() == '':
            errors['name'].append('Name cannot be empty')
            have_errors = True
        try:
            hasattr(AcctType, data['acct_type'])
        except Exception:
            errors['acct_type'].append('"%s" is not a valid Account Type',
                                       data['acct_type'])
        if data['credit_limit'].strip() != '':
            errors = self._validate_decimal('credit_limit', data, errors)
        if data['apr'].strip() != '':
            errors = self._validate_decimal('apr', data, errors)
        if data['ofxgetter_config_json'].strip() != '':
            try:
                json.loads(data['ofxgetter_config_json'])
            except Exception:
                errors['ofxgetter_config_json'].append('Invalid JSON!')
        if data['prime_rate_margin'].strip() != '':
            errors = self._validate_decimal('prime_rate_margin', data, errors)
        if data['interest_class_name'] not in INTEREST_CALCULATION_NAMES:
            errors['interest_class_name'].append('Invalid interest class')
        if data['min_payment_class_name'] not in MIN_PAYMENT_FORMULA_NAMES:
            errors['min_payment_class_name'].append(
                'Invalid minimum payment class name'
            )
        if data['plaid_account'] != 'null,null':
            iid, aid = data['plaid_account'].split(',')
            acctid = 0
            if 'id' in data and data['id'].strip() != '':
                acctid = int(data['id'])
            existing = db_session.query(Account).filter(
                Account.plaid_item_id == self.fix_string(iid),
                Account.plaid_account_id == self.fix_string(aid),
                Account.id != acctid
            ).all()
            if existing:
                errors['plaid_account'].append(
                    'ERROR: This Plaid account is already used by the Account: '
                    f'{existing[0].name} ({existing[0].id}). Please remove it '
                    f'from that account before adding it to another.'
                )
        for f in RE_FIELD_NAMES:
            if data[f].strip() == '':
                continue
            try:
                re.compile(data[f])
            except Exception:
                errors[f].append('Invalid regular expression.')
        if have_errors:
            return errors
        for k, v in errors.items():
            if v:
                return errors
        return None

    def submit(self, data):
        """
        Handle form submission; create or update models in the DB. Raises an
        Exception for any errors.

        :param data: submitted form data
        :type data: dict
        :return: message describing changes to DB (i.e. link to created record)
        :rtype: str
        """
        new_acct = False
        if 'id' in data and data['id'].strip() != '':
            # updating an existing account
            account = db_session.query(Account).get(int(data['id']))
            if account is None:
                raise RuntimeError("Error: no Account with ID %s" % data['id'])
            action = 'updating Account ' + data['id']
        else:
            new_acct = True
            account = Account()
            action = 'creating new Account'
        account.name = data['name'].strip()
        account.description = self.fix_string(data['description'])
        account.acct_type = getattr(AcctType, data['acct_type'])
        account.ofx_cat_memo_to_name = data['ofx_cat_memo_to_name']
        account.vault_creds_path = self.fix_string(data['vault_creds_path'])
        account.ofxgetter_config_json = self.fix_string(
            data['ofxgetter_config_json']
        )
        account.negate_ofx_amounts = data['negate_ofx_amounts']
        account.reconcile_trans = data['reconcile_trans']
        if account.acct_type == AcctType.Credit:
            if data['credit_limit'].strip() != '':
                account.credit_limit = Decimal(data['credit_limit'])
            else:
                account.credit_limit = None
            if data['apr'].strip() != '':
                account.apr = Decimal(data['apr'])
                if account.apr > Decimal('1'):
                    account.apr = account.apr * Decimal('0.01')
            else:
                account.apr = None
            if data['prime_rate_margin'].strip() != '':
                account.prime_rate_margin = Decimal(data['prime_rate_margin'])
                if account.prime_rate_margin > Decimal('1'):
                    account.prime_rate_margin = account.prime_rate_margin * \
                                                Decimal('0.01')
            else:
                account.prime_rate_margin = None
            account.interest_class_name = data['interest_class_name']
            account.min_payment_class_name = data['min_payment_class_name']
        account.is_active = data['is_active']
        for f in RE_FIELD_NAMES:
            data[f] = data[f].strip()
            if data[f] == '':
                data[f] = None
            setattr(account, f, data[f])
        if data['plaid_account'] == 'null,null':
            account.plaid_item_id = None
            account.plaid_account_id = None
        else:
            iid, aid = data['plaid_account'].split(',')
            account.plaid_item_id = self.fix_string(iid)
            account.plaid_account_id = self.fix_string(aid)
        logger.info('%s: %s', action, account.as_dict)
        db_session.add(account)
        db_session.commit()
        if new_acct:
            account.set_balance(ledger=Decimal('0'), avail=Decimal('0'))
            db_session.add(account)
            db_session.commit()
        return 'Successfully saved Account %d in database.' % account.id


class AccountTxfrFormHandler(FormHandlerView):
    """
    Handle POST /forms/account_transfer
    """

    def validate(self, data):
        """
        Validate the form data. Return None if it is valid, or else a hash of
        field names to list of error strings for each field.

        :param data: submitted form data
        :type data: dict
        :return: None if no errors, or hash of field name to errors for that
          field
        """
        have_errors = False
        errors = {k: [] for k in data.keys()}
        if data['date'].strip() == '':
            errors['date'].append('Transactions must have a date')
            have_errors = True
        else:
            try:
                datetime.strptime(data['date'], '%Y-%m-%d').date()
            except Exception:
                errors['date'].append(
                    'Date "%s" is not valid (YYYY-MM-DD)' % data['date']
                )
                have_errors = True
        if float(data['amount']) == 0:
            errors['amount'].append('Amount cannot be zero')
            have_errors = True
        if float(data['amount']) < 0:
            errors['amount'].append('Amount cannot be negative')
            have_errors = True
        if data['budget'] == 'None':
            errors['budget'].append('Transactions must have a budget')
            have_errors = True
        if data['from_account'] == 'None':
            errors['from_account'].append('From Account cannot be empty')
            have_errors = True
        else:
            from_acct = db_session.query(Account).get(int(data['from_account']))
            if from_acct is None:
                errors['from_account'].append('from_account ID does not exist')
                have_errors = True
            else:
                if not from_acct.is_active:
                    errors['from_account'].append('From Account must be active')
                    have_errors = True
                if from_acct.acct_type not in AcctType.transferrable_types():
                    errors['from_account'].append(
                        'From Account type is not transferrable'
                    )
                    have_errors = True
        if data['to_account'] == 'None':
            errors['to_account'].append('To Account cannot be empty')
            have_errors = True
        else:
            to_acct = db_session.query(Account).get(int(data['to_account']))
            if to_acct is None:
                errors['to_account'].append('to_account ID does not exist')
                have_errors = True
            else:
                if not to_acct.is_active:
                    errors['to_account'].append('To Account must be active')
                    have_errors = True
                if to_acct.acct_type not in AcctType.transferrable_types():
                    errors['to_account'].append(
                        'To Account type is not transferrable'
                    )
                    have_errors = True
        if have_errors:
            return errors
        return None

    def submit(self, data):
        """
        Handle form submission; create or update models in the DB. Raises an
        Exception for any errors.

        :param data: submitted form data
        :type data: dict
        :return: message describing changes to DB (i.e. link to created record)
        :rtype: str
        """
        # get the data
        trans_date = datetime.strptime(data['date'], '%Y-%m-%d').date()
        amt = Decimal(data['amount'])
        from_acct = db_session.query(Account).get(int(data['from_account']))
        if from_acct is None:
            raise RuntimeError(
                "Error: no Account with ID %s" % data['from_account']
            )
        to_acct = db_session.query(Account).get(int(data['to_account']))
        if to_acct is None:
            raise RuntimeError(
                "Error: no Account with ID %s" % data['to_account']
            )
        budget = db_session.query(Budget).get(int(data['budget']))
        if budget is None:
            raise RuntimeError(
                "Error: no Budget with ID %s" % data['budget']
            )
        notes = data['notes'].strip()
        desc = 'Account Transfer - %s from %s (%d) to %s (%d)' % (
            amt, from_acct.name, from_acct.id, to_acct.name, to_acct.id
        )
        logger.info(desc)
        t1 = Transaction(
            date=trans_date,
            budget_amounts={budget: amt},
            budgeted_amount=amt,
            description=desc,
            account=from_acct,
            notes=notes,
            planned_budget=budget
        )
        db_session.add(t1)
        t2 = Transaction(
            date=trans_date,
            budget_amounts={budget: (-1 * amt)},
            budgeted_amount=(-1 * amt),
            description=desc,
            account=to_acct,
            notes=notes,
            planned_budget=budget
        )
        db_session.add(t2)
        t1.transfer = t2
        db_session.add(t1)
        t2.transfer = t1
        db_session.add(t2)
        db_session.commit()
        return 'Successfully saved Transactions %d and %d in database.' % (
            t1.id, t2.id
        )


app.add_url_rule('/accounts', view_func=AccountsView.as_view('accounts_view'))
app.add_url_rule(
    '/ajax/account/<int:account_id>',
    view_func=AccountAjax.as_view('account_ajax')
)
app.add_url_rule(
    '/accounts/<int:acct_id>',
    view_func=OneAccountView.as_view('account_view')
)
app.add_url_rule(
    '/forms/account',
    view_func=AccountFormHandler.as_view('account_form')
)
app.add_url_rule(
    '/forms/account_transfer',
    view_func=AccountTxfrFormHandler.as_view('account_transfer_form')
)



"""
The latest version of this package is available at:
<http://github.com/jantman/biweeklybudget>

################################################################################
Copyright 2016-2024 Jason Antman <http://www.jasonantman.com>

    This file is part of biweeklybudget, also known as biweeklybudget.

    biweeklybudget is free software: you can redistribute it and/or modify
    it under the terms of the GNU Affero General Public License as published by
    the Free Software Foundation, either version 3 of the License, or
    (at your option) any later version.

    biweeklybudget is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU Affero General Public License for more details.

    You should have received a copy of the GNU Affero General Public License
    along with biweeklybudget.  If not, see <http://www.gnu.org/licenses/>.

The Copyright and Authors attributions contained herein may not be removed or
otherwise altered, except to add the Author attribution of a contributor to
this work. (Additional Terms pursuant to Section 7b of the AGPL v3)
################################################################################
While not legally required, I sincerely request that anyone who finds
bugs please submit them at <https://github.com/jantman/biweeklybudget> or
to me via email, and that you send any contributions or improvements
either as a pull request on GitHub, or to me via email.
################################################################################

AUTHORS:
Jason Antman <jason@jasonantman.com> <http://www.jasonantman.com>
################################################################################

This is mostly based on http://flask.pocoo.org/docs/0.12/patterns/sqlalchemy/
"""

import logging
import time
import os
from sqlalchemy import event, inspect

from biweeklybudget.models.account import Account
from biweeklybudget.models.budget_model import Budget
from biweeklybudget.models.budget_transaction import BudgetTransaction
from biweeklybudget.models.ofx_transaction import OFXTransaction
from biweeklybudget.utils import fmt_currency

logger = logging.getLogger(__name__)


def handle_budget_trans_amount_change(**kwargs):
    """
    Handle change of :py:attr:`.BudgetTransaction.amount` for existing
    instances (``trans_id`` is not None). For new or deleted instances, we rely
    on :py:func:`~.handle_new_or_deleted_budget_transaction` called via
    :py:func:`~.handle_before_flush`.

    If the BudgetTransaction's :py:attr:`~.BudgetTransaction.budget` uses a
    :py:class:`~.Budget` with :py:attr:`~.Budget.is_periodic` ``False`` (i.e. a
    standing budget), update the Budget's :py:attr:`~.Budget.current_balance`
    for this transaction.

    See: :py:meth:`sqlalchemy.orm.events.AttributeEvents.set`

    :param kwargs: keyword arguments
    :type kwargs: dict
    """
    tgt = kwargs['target']
    if tgt.trans_id is None:
        logger.debug('got BudgetTransaction with trans_id None; skipping')
        return
    if tgt.budget.is_periodic:
        logger.debug('got BudgetTransaction with periodic budget; skipping')
        return
    value = kwargs['value']
    oldvalue = kwargs['oldvalue']
    session = inspect(tgt).session
    diff = oldvalue - value
    old_budg_curr = tgt.budget.current_balance
    new_budg = old_budg_curr + diff
    logger.info(
        'Handle BudgetTransaction %d against standing budget %d UPDATE; '
        'actual_amount change from %s to %s; update budget current_balance '
        'from %s to %s',
        tgt.id, tgt.budget.id, oldvalue, value, old_budg_curr, new_budg
    )
    tgt.budget.current_balance = new_budg
    session.add(tgt.budget)


def handle_new_or_deleted_budget_transaction(session):
    """
    ``before_flush`` event handler
    (:py:meth:`sqlalchemy.orm.events.SessionEvents.before_flush`)
    on the DB session, to handle creation of *new* BudgetTransactions or
    deletion of BudgetTransactions. For updates to existing BudgetTransactions,
    we rely on :py:func:`~.handle_budget_trans_amount_change`.

    If the BudgetTransaction's :py:attr:`~.BudgetTransaction.budget` is a
    :py:class:`~.Budget` with :py:attr:`~.Budget.is_periodic` ``False`` (i.e. a
    standing budget), update the Budget's :py:attr:`~.Budget.current_balance`
    for this transaction.

    :param session: current database session
    :type session: sqlalchemy.orm.session.Session
    """
    # handle NEW
    updated = 0
    for obj in session.new:
        if not isinstance(obj, BudgetTransaction):
            continue
        if obj.budget is not None:
            budg = obj.budget
        else:
            budg = session.query(Budget).get(obj.budget_id)
        if budg.is_periodic:
            continue
        logger.debug(
            'Session has new BudgetTransaction referencing standing '
            'budget id=%s', budg.id
        )
        old_amt = budg.current_balance
        budg.current_balance = old_amt - obj.amount
        logger.info(
            'New BudgetTransaction (%s) for %s against standing budget id=%s; '
            'update budget current_balance from %s to %s', obj,
            fmt_currency(obj.amount), budg.id, fmt_currency(old_amt),
            fmt_currency(budg.current_balance)
        )
        session.add(budg)
        updated += 1
    logger.debug(
        'Done handling new BudgetTransactions; updated %d standing budgets',
        updated
    )
    # handle DELETED
    updated = 0
    for obj in session.deleted:
        if not isinstance(obj, BudgetTransaction):
            continue
        if obj.budget is not None:
            budg = obj.budget
        else:
            budg = session.query(Budget).get(obj.budget_id)
        if budg.is_periodic:
            continue
        logger.debug(
            'Session has deleted BudgetTransaction referencing standing '
            'budget id=%s', budg.id
        )
        old_amt = budg.current_balance
        budg.current_balance = old_amt + obj.amount
        logger.info(
            'Deleted BudgetTransaction (%s) for %s against standing budget '
            'id=%s; update budget current_balance from %s to %s',
            obj, fmt_currency(obj.amount), budg.id,
            fmt_currency(old_amt), fmt_currency(budg.current_balance)
        )
        session.add(budg)
        updated += 1
    logger.debug(
        'Done handling deleted BudgetTransactions; '
        'updated %d standing budgets', updated
    )


def handle_ofx_transaction_new_or_change(session):
    """
    ``before_flush`` event handler
    (:py:meth:`sqlalchemy.orm.events.SessionEvents.before_flush`)
    on the DB session, to handle setting the ``is_*`` fields on new or changed
    OFXTransaction instances according to its Account.

    :param session: current database session
    :type session: sqlalchemy.orm.session.Session
    """
    for obj in session.new:
        if isinstance(obj, OFXTransaction):
            try:
                obj.update_is_fields()
            except Exception:
                logger.error('Error setting OFXTransaction is_ fields',
                             exc_info=True)
    for obj in session.dirty:
        if isinstance(obj, OFXTransaction):
            try:
                obj.update_is_fields()
            except Exception:
                logger.error('Error setting OFXTransaction is_ fields',
                             exc_info=True)


def handle_account_re_change(session):
    """
    Handler for change of one of:

    * :py:attr:`~.Account.re_interest_paid`
    * :py:attr:`~.Account.re_interest_charge`
    * :py:attr:`~.Account.re_late_fee`
    * :py:attr:`~.Account.re_other_fee`
    * :py:attr:`~.Account.re_payment`

    When one of these regexes is changed on an Account, we trigger a re-run
    of :py:meth:`~.OFXTransaction.update_is_fields` on all OFXTransactions for
    the account.

    :param session: current database session
    :type session: sqlalchemy.orm.session.Session
    """
    attrs = [
        're_interest_paid',
        're_interest_charge',
        're_late_fee',
        're_other_fee',
        're_payment'
    ]
    for obj in session.dirty:
        if not isinstance(obj, Account):
            continue
        changed = []
        insp = inspect(obj)
        for attr in attrs:
            hx = getattr(insp.attrs, attr).history
            if hx is None or hx.added is None or hx.deleted is None:
                continue
            if len(hx.added) > 0 and len(hx.deleted) > 0:
                logger.debug(
                    '%s %s changed from %s to %s',
                    obj, attr, hx.deleted, hx.added
                )
                changed.append(attr)
        if len(changed) < 1:
            continue
        logger.debug(
            '%s has regex changes; triggering update_is_fields() on all child '
            'OFXTransactions.', obj
        )
        for stmt in obj.all_statements:
            for txn in stmt.ofx_trans:
                txn.update_is_fields()
        logger.debug('Done with update_is_fields() for %s', obj)


def handle_before_flush(session, flush_context, instances):
    """
    Hook into ``before_flush``
    (:py:meth:`sqlalchemy.orm.events.SessionEvents.before_flush`)
    on the DB session, to handle updates that need to be made before persisting
    data. Currently, this method just calls a number of other methods to handle
    specific cases:

    * :py:func:`~.handle_new_or_deleted_budget_transaction`

    :param session: current database session
    :type session: sqlalchemy.orm.session.Session
    :param flush_context: internal SQLAlchemy object
    :type flush_context: sqlalchemy.orm.session.UOWTransaction
    :param instances: deprecated
    """
    logger.debug('handle_before_flush handler')
    handle_new_or_deleted_budget_transaction(session)
    handle_ofx_transaction_new_or_change(session)
    handle_account_re_change(session)
    logger.debug('handle_before_flush done')


def query_profile_before(conn, cursor, statement, parameters, context, _):  # noqa
    """
    Query profiling database event listener, to be added as listener on the
    Engine's ``before_cursor_execute`` event.

    For information, see:
    http://docs.sqlalchemy.org/en/latest/faq/performance.html#query-profiling
    """
    conn.info.setdefault('query_start_time', []).append(time.time())
    logger.debug(
        "Start Query: %s; parameters: %s", statement.replace('\n', ' '),
        parameters
    )


def query_profile_after(conn, cursor, statement, parameters, context, _):  # noqa
    """
    Query profiling database event listener, to be added as listener on the
    Engine's ``after_cursor_execute`` event.

    For information, see:
    http://docs.sqlalchemy.org/en/latest/faq/performance.html#query-profiling
    """
    total = time.time() - conn.info['query_start_time'].pop(-1)
    logger.debug(
        "Query complete in %f seconds. Query: %s; parameters: %s", total,
        statement.replace('\n', ' '), parameters
    )


def init_event_listeners(db_session, engine):
    """
    Initialize/register all SQLAlchemy event listeners.

    See http://docs.sqlalchemy.org/en/latest/orm/events.html

    :param db_session: the Database Session
    :type db_session: sqlalchemy.orm.scoping.scoped_session
    :param engine: top-level Database Engine instance
    :type engine: sqlalchemy.engine.Engine
    """
    if os.environ.get('SQL_QUERY_PROFILE', 'false') == 'true':
        logger.debug('Enabling SQL query timing event handlers.')
        event.listen(engine, 'before_cursor_execute', query_profile_before)
        event.listen(engine, 'after_cursor_execute', query_profile_after)
    logger.debug('Setting up DB model event listeners')
    event.listen(
        BudgetTransaction.amount,
        'set',
        handle_budget_trans_amount_change,
        active_history=True,
        named=True
    )
    event.listen(
        db_session,
        'before_flush',
        handle_before_flush
    )



"""
The latest version of this package is available at:
<http://github.com/jantman/biweeklybudget>

################################################################################
Copyright 2016-2024 Jason Antman <http://www.jasonantman.com>

    This file is part of biweeklybudget, also known as biweeklybudget.

    biweeklybudget is free software: you can redistribute it and/or modify
    it under the terms of the GNU Affero General Public License as published by
    the Free Software Foundation, either version 3 of the License, or
    (at your option) any later version.

    biweeklybudget is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU Affero General Public License for more details.

    You should have received a copy of the GNU Affero General Public License
    along with biweeklybudget.  If not, see <http://www.gnu.org/licenses/>.

The Copyright and Authors attributions contained herein may not be removed or
otherwise altered, except to add the Author attribution of a contributor to
this work. (Additional Terms pursuant to Section 7b of the AGPL v3)
################################################################################
While not legally required, I sincerely request that anyone who finds
bugs please submit them at <https://github.com/jantman/biweeklybudget> or
to me via email, and that you send any contributions or improvements
either as a pull request on GitHub, or to me via email.
################################################################################

AUTHORS:
Jason Antman <jason@jasonantman.com> <http://www.jasonantman.com>
################################################################################
"""

from datetime import datetime
import logging
from decimal import Decimal

from flask.views import MethodView
from flask import render_template, request, redirect

from biweeklybudget.flaskapp.app import app
from biweeklybudget.utils import dtnow
from biweeklybudget.biweeklypayperiod import BiweeklyPayPeriod
from biweeklybudget.models.budget_model import Budget
from biweeklybudget.models.account import Account
from biweeklybudget.models.scheduled_transaction import ScheduledTransaction
from biweeklybudget.models.transaction import Transaction
from biweeklybudget.models.txn_reconcile import TxnReconcile
from biweeklybudget.db import db_session
from biweeklybudget.flaskapp.views.formhandlerview import FormHandlerView

logger = logging.getLogger(__name__)


class PayPeriodsView(MethodView):
    """
    Render the top-level GET /payperiods view using ``payperiods.html`` template
    """

    def get(self):
        pp = BiweeklyPayPeriod.period_for_date(dtnow(), db_session)
        pp_curr_idx = 1
        pp_next_idx = 2
        pp_following_idx = 3
        periods = [
            pp.previous,
            pp
        ]
        x = pp
        # add another
        for i in range(0, 8):
            x = x.next
            periods.append(x)
        # trigger calculation/cache of data before passing on to jinja
        for p in periods:
            p.overall_sums
        return render_template(
            'payperiods.html',
            periods=periods,
            curr_pp=pp,
            pp_curr_idx=pp_curr_idx,
            pp_next_idx=pp_next_idx,
            pp_following_idx=pp_following_idx
        )


class PayPeriodView(MethodView):
    """
    Render the single PayPeriod GET /payperiod/YYYY-MM-DD view using the
    ``payperiod.html`` template.
    """

    def suffix_for_period(self, curr_pp, pp):
        """
        Generate the suffix to use for the given pay period in the view

        :param curr_pp: the current (today) pay period
        :type curr_pp: BiweeklyPayPeriod
        :param pp: the pay period in question
        :type pp: BiweeklyPayPeriod
        :return: suffix for the pay period
        :rtype: str
        """
        if pp.start_date == curr_pp.start_date:
            return '(curr.)'
        if pp.start_date == curr_pp.next.start_date:
            return '(next)'
        if pp.start_date == curr_pp.previous.start_date:
            return '(prev.)'
        return ''

    def get(self, period_date):
        d = datetime.strptime(period_date, '%Y-%m-%d').date()
        pp = BiweeklyPayPeriod.period_for_date(d, db_session)
        curr_pp = BiweeklyPayPeriod.period_for_date(dtnow(), db_session)
        budgets = {}
        active_budgets = {}
        for b in db_session.query(Budget).all():
            k = b.name
            if b.is_income:
                k = '%s (i)' % b.name
            budgets[b.id] = k
            if b.is_active:
                active_budgets[b.id] = k
        standing = {
            b.id: b.current_balance
            for b in db_session.query(Budget).filter(
                Budget.is_periodic.__eq__(False),
                Budget.is_active.__eq__(True)
            ).all()
        }
        periodic = {
            b.id: b.current_balance
            for b in db_session.query(Budget).filter(
                Budget.is_periodic.__eq__(True),
                Budget.is_active.__eq__(True)
            ).all()
        }
        # Issue #161 - backfill inactive budgets that have transactions
        # in this payperiod.
        for budg_id in pp.budget_sums:
            if budg_id not in periodic:
                b = db_session.query(Budget).get(budg_id)
                periodic[b.id] = b.current_balance
        accts = {a.name: a.id for a in db_session.query(Account).all()}
        txfr_date_str = dtnow().strftime('%Y-%m-%d')
        if dtnow().date() < pp.start_date or dtnow().date() > pp.end_date:
            # If we're looking at a non-current pay period, default the
            # transfer modal date to the start of the period.
            txfr_date_str = pp.start_date.strftime('%Y-%m-%d')
        return render_template(
            'payperiod.html',
            pp=pp,
            pp_prev_date=pp.previous.start_date,
            pp_prev_sums=pp.previous.overall_sums,
            pp_prev_suffix=self.suffix_for_period(curr_pp, pp.previous),
            pp_curr_date=pp.start_date,
            pp_curr_sums=pp.overall_sums,
            pp_curr_suffix=self.suffix_for_period(curr_pp, pp),
            pp_next_date=pp.next.start_date,
            pp_next_sums=pp.next.overall_sums,
            pp_next_suffix=self.suffix_for_period(curr_pp, pp.next),
            pp_following_date=pp.next.next.start_date,
            pp_following_sums=pp.next.next.overall_sums,
            pp_following_suffix=self.suffix_for_period(curr_pp, pp.next.next),
            pp_last_date=pp.next.next.next.start_date,
            pp_last_sums=pp.next.next.next.overall_sums,
            pp_last_suffix=self.suffix_for_period(curr_pp, pp.next.next.next),
            budget_sums=pp.budget_sums,
            budgets=budgets,
            standing=standing,
            periodic=periodic,
            transactions=pp.transactions_list,
            accts=accts,
            txfr_date_str=txfr_date_str,
            active_budgets=active_budgets
        )


class PeriodForDateView(MethodView):
    """
    Render a redirect from a given date to the pay period for that date
    """

    def get(self):
        d_str = request.args.get('date', None)
        if d_str is None:
            pp = BiweeklyPayPeriod.period_for_date(
                dtnow().date(), db_session
            )
            logger.debug('Redirect to current payperiod: %s', pp)
        else:
            d = datetime.strptime(d_str, '%Y-%m-%d').date()
            pp = BiweeklyPayPeriod.period_for_date(d, db_session)
            logger.debug('Found period for %s (%s): %s', d_str, d, pp)
        return redirect(
            '/payperiod/%s' % pp.start_date.strftime('%Y-%m-%d'),
            code=302
        )


class SchedToTransFormHandler(FormHandlerView):
    """
    Handle POST /forms/sched_to_trans
    """

    def validate(self, data):
        """
        Validate the form data. Return None if it is valid, or else a hash of
        field names to list of error strings for each field.

        :param data: submitted form data
        :type data: dict
        :return: None if no errors, or hash of field name to errors for that
          field
        """
        _id = int(data['id'])
        # make sure the ID is valid
        db_session.query(ScheduledTransaction).get(_id)
        d = datetime.strptime(data['date'], '%Y-%m-%d').date()
        pp = BiweeklyPayPeriod.period_for_date(d, db_session)
        have_errors = False
        errors = {k: [] for k in data.keys()}
        if data.get('description', '').strip() == '':
            errors['description'].append('Description cannot be empty')
            have_errors = True
        if float(data['amount']) == 0:
            errors['amount'].append('Amount cannot be zero')
            have_errors = True
        if d < pp.start_date or d > pp.end_date:
            errors['date'].append('Date must be in current pay period')
            have_errors = True
        if have_errors:
            return errors
        return None

    def submit(self, data):
        """
        Handle form submission; create or update models in the DB. Raises an
        Exception for any errors.

        :param data: submitted form data
        :type data: dict
        :return: message describing changes to DB (i.e. link to created record)
        :rtype: str
        """
        st_id = int(data['id'])
        st = db_session.query(ScheduledTransaction).get(st_id)
        d = datetime.strptime(data['date'], '%Y-%m-%d').date()
        if data.get('sales_tax', '').strip() == '':
            sales_tax = Decimal('0.0')
        else:
            sales_tax = Decimal(data['sales_tax'])
        t = Transaction(
            date=d,
            budget_amounts={st.budget: Decimal(data['amount'])},
            budgeted_amount=st.amount,
            sales_tax=sales_tax,
            description=data['description'],
            notes=data['notes'],
            account=st.account,
            planned_budget=st.budget,
            scheduled_trans=st
        )
        db_session.add(t)
        db_session.commit()
        logger.info('Created Transaction %d for ScheduledTransaction %d',
                    t.id, st.id)
        return 'Successfully created Transaction %d ' \
               'for ScheduledTransaction %d.' % (t.id, st.id)


class SkipSchedTransFormHandler(FormHandlerView):
    """
    Handle POST /forms/skip_sched_trans
    """

    def validate(self, data):
        """
        Validate the form data. Return None if it is valid, or else a hash of
        field names to list of error strings for each field.

        :param data: submitted form data
        :type data: dict
        :return: None if no errors, or hash of field name to errors for that
          field
        """
        _id = int(data['id'])
        # make sure the ID is valid
        db_session.query(ScheduledTransaction).get(_id)
        d = datetime.strptime(data['payperiod_start_date'], '%Y-%m-%d').date()
        BiweeklyPayPeriod.period_for_date(d, db_session)
        have_errors = False
        errors = {k: [] for k in data.keys()}
        if data.get('notes', '').strip() == '':
            errors['notes'].append('Notes cannot be empty')
            have_errors = True
        if have_errors:
            return errors
        return None

    def submit(self, data):
        """
        Handle form submission; create or update models in the DB. Raises an
        Exception for any errors.

        :param data: submitted form data
        :type data: dict
        :return: message describing changes to DB (i.e. link to created record)
        :rtype: str
        """
        st_id = int(data['id'])
        st = db_session.query(ScheduledTransaction).get(st_id)
        d = datetime.strptime(data['payperiod_start_date'], '%Y-%m-%d').date()
        desc = 'Skip ScheduledTransaction %d in period %s' % (
            st_id, data['payperiod_start_date']
        )
        t = Transaction(
            date=d,
            budget_amounts={st.budget: Decimal('0.0')},
            budgeted_amount=Decimal('0.0'),
            description=desc,
            notes=data['notes'],
            account=st.account,
            scheduled_trans=st,
            planned_budget=st.budget
        )
        db_session.add(t)
        db_session.add(TxnReconcile(
            transaction=t,
            note=desc
        ))
        db_session.commit()
        logger.info('Created Transaction %d to skip '
                    'ScheduledTransaction %d', t.id, st.id)
        return 'Successfully created Transaction %d to skip ' \
               'ScheduledTransaction %d.' % (t.id, st.id)


app.add_url_rule(
    '/payperiods',
    view_func=PayPeriodsView.as_view('payperiods_view')
)

app.add_url_rule(
    '/payperiod/<period_date>',
    view_func=PayPeriodView.as_view('payperiod_view')
)

app.add_url_rule(
    '/pay_period_for',
    view_func=PeriodForDateView.as_view('pay_period_for_view')
)

app.add_url_rule(
    '/forms/sched_to_trans',
    view_func=SchedToTransFormHandler.as_view('sched_to_trans_form')
)

app.add_url_rule(
    '/forms/skip_sched_trans',
    view_func=SkipSchedTransFormHandler.as_view('skip_sched_trans_form')
)



"""
The latest version of this package is available at:
<http://github.com/jantman/biweeklybudget>

################################################################################
Copyright 2016-2024 Jason Antman <http://www.jasonantman.com>

    This file is part of biweeklybudget, also known as biweeklybudget.

    biweeklybudget is free software: you can redistribute it and/or modify
    it under the terms of the GNU Affero General Public License as published by
    the Free Software Foundation, either version 3 of the License, or
    (at your option) any later version.

    biweeklybudget is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU Affero General Public License for more details.

    You should have received a copy of the GNU Affero General Public License
    along with biweeklybudget.  If not, see <http://www.gnu.org/licenses/>.

The Copyright and Authors attributions contained herein may not be removed or
otherwise altered, except to add the Author attribution of a contributor to
this work. (Additional Terms pursuant to Section 7b of the AGPL v3)
################################################################################
While not legally required, I sincerely request that anyone who finds
bugs please submit them at <https://github.com/jantman/biweeklybudget> or
to me via email, and that you send any contributions or improvements
either as a pull request on GitHub, or to me via email.
################################################################################

AUTHORS:
Jason Antman <jason@jasonantman.com> <http://www.jasonantman.com>
################################################################################
"""

import logging

from flask.views import MethodView
from flask import render_template, jsonify, request

from biweeklybudget.flaskapp.app import app
from biweeklybudget.models.budget_model import Budget
from biweeklybudget.models.account import Account
from biweeklybudget.models.txn_reconcile import TxnReconcile
from biweeklybudget.models.transaction import Transaction
from biweeklybudget.models.ofx_transaction import OFXTransaction
from biweeklybudget.db import db_session

logger = logging.getLogger(__name__)


class ReconcileView(MethodView):
    """
    Render the top-level GET /reconcile view using ``reconcile.html`` template.
    """

    def get(self):
        budgets = {}
        active_budgets = {}
        for b in db_session.query(Budget).all():
            k = b.name
            if b.is_income:
                k = '%s (i)' % b.name
            budgets[b.id] = k
            if b.is_active:
                active_budgets[b.id] = k
        accts = {a.name: a.id for a in db_session.query(Account).all()}
        return render_template(
            'reconcile.html',
            budgets=budgets,
            accts=accts,
            active_budgets=active_budgets
        )


class TxnReconcileAjax(MethodView):
    """
    Handle GET /ajax/reconcile/<int:reconcile_id> endpoint.
    """

    def get(self, reconcile_id):
        rec = db_session.query(TxnReconcile).get(reconcile_id)

        res = {
            'reconcile': rec.as_dict,
            'transaction': rec.transaction.as_dict
        }
        res['transaction']['budgets'] = [
            {
                'name': bt.budget.name,
                'id': bt.budget_id,
                'amount': bt.amount,
                'is_income': bt.budget.is_income
            }
            for bt in sorted(
                rec.transaction.budget_transactions,
                key=lambda x: x.amount, reverse=True
            )
        ]
        if rec.ofx_trans is not None:
            res['ofx_trans'] = rec.ofx_trans.as_dict
            res['ofx_stmt'] = rec.ofx_trans.statement.as_dict
            res['acct_id'] = rec.ofx_trans.account_id
            res['acct_name'] = rec.ofx_trans.account.name
        else:
            res['ofx_trans'] = None
            res['ofx_stmt'] = None
            res['acct_id'] = rec.transaction.account_id
            res['acct_name'] = rec.transaction.account.name
        return jsonify(res)


class OfxUnreconciledAjax(MethodView):
    """
    Handle GET /ajax/unreconciled/ofx endpoint.
    """

    def get(self):
        res = []
        for t in OFXTransaction.unreconciled(
            db_session
        ).order_by(OFXTransaction.date_posted, OFXTransaction.fitid).all():
            d = t.as_dict
            d['account_name'] = t.account.name
            d['account_amount'] = t.account_amount
            res.append(d)
        return jsonify(res)


class TransUnreconciledAjax(MethodView):
    """
    Handle GET /ajax/unreconciled/trans endpoint.
    """

    def get(self):
        res = []
        for t in Transaction.unreconciled(
            db_session
        ).order_by(Transaction.date, Transaction.id).all():
            d = t.as_dict
            d['budgets'] = [
                {
                    'name': bt.budget.name,
                    'id': bt.budget_id,
                    'amount': bt.amount,
                    'is_income': bt.budget.is_income
                }
                for bt in sorted(
                    t.budget_transactions, key=lambda x: x.amount,
                    reverse=True
                )
            ]
            d['account_name'] = t.account.name
            res.append(d)
        return jsonify(res)


class ReconcileAjax(MethodView):
    """
    Handle POST ``/ajax/reconcile`` endpoint.
    """

    def post(self):
        """
        Handle POST ``/ajax/reconcile``

        Request is a JSON dict with two keys, "reconciled" and "ofxIgnored".
        "reconciled" value is a dict of integer transaction ID keys, to
        values which are either a string reason why the Transaction is being
        reconciled as "No OFX" or a 2-item list of OFXTransaction acct_id
        and fitid.
        "ofxIgnored" is a dict with string keys which are strings identifying
        an OFXTransaction in the form "<ACCT_ID>%<FITID>", and values are a
        string reason why the OFXTransaction is being reconciled without a
        matching Transaction.

        Response is a JSON dict. Keys are ``success`` (boolean) and either
        ``error_message`` (string) or ``success_message`` (string).

        :return: JSON response
        """
        raw = request.get_json()
        data = {
            'reconciled': {
                int(x): raw['reconciled'][x] for x in raw['reconciled']
            },
            'ofxIgnored': raw.get('ofxIgnored', {})
        }
        logger.debug('POST /ajax/reconcile: %s', data)
        rec_count = 0
        for trans_id in sorted(data['reconciled'].keys()):
            trans = db_session.query(Transaction).get(trans_id)
            if trans is None:
                logger.error('Invalid transaction ID: %s', trans_id)
                return jsonify({
                    'success': False,
                    'error_message': 'Invalid Transaction ID: %s' % trans_id
                }), 400
            if not isinstance(data['reconciled'][trans_id], type([])):
                # it's a string; reconcile without OFX
                db_session.add(TxnReconcile(
                    txn_id=trans_id,
                    note=data['reconciled'][trans_id]
                ))
                logger.info(
                    'Reconcile %s as NoOFX; note=%s',
                    trans, data['reconciled'][trans_id]
                )
                rec_count += 1
                continue
            # else reconcile with OFX
            ofx_key = (
                data['reconciled'][trans_id][0], data['reconciled'][trans_id][1]
            )
            ofx = db_session.query(OFXTransaction).get(ofx_key)
            if ofx is None:
                logger.error('Invalid OFXTransaction: %s', ofx_key)
                return jsonify({
                    'success': False,
                    'error_message': 'Invalid OFXTransaction: (%s, \'%s\')' % (
                        ofx_key[0], ofx_key[1]
                    )
                }), 400
            db_session.add(TxnReconcile(
                txn_id=trans_id,
                ofx_account_id=data['reconciled'][trans_id][0],
                ofx_fitid=data['reconciled'][trans_id][1]
            ))
            logger.info('Reconcile %s with %s', trans, ofx)
            rec_count += 1
        # handle OFXTransactions to reconcile with no Transaction
        for ofxkey in sorted(data['ofxIgnored'].keys()):
            note = data['ofxIgnored'][ofxkey]
            acct_id, fitid = ofxkey.split('%', 1)
            db_session.add(TxnReconcile(
                ofx_account_id=acct_id,
                ofx_fitid=fitid,
                note=note
            ))
            logger.info(
                'Reconcile OFXTransaction (%s, %s) as NoTransaction; note=%s',
                acct_id, fitid, note
            )
            rec_count += 1
        try:
            db_session.flush()
            db_session.commit()
        except Exception as ex:
            logger.error('Exception committing transaction reconcile',
                         exc_info=True)
            return jsonify({
                'success': False,
                'error_message': 'Exception committing reconcile(s): %s' % ex
            }), 400
        return jsonify({
            'success': True,
            'success_message': 'Successfully reconciled '
                               '%d transactions' % rec_count
        })


app.add_url_rule(
    '/reconcile',
    view_func=ReconcileView.as_view('reconcile_view')
)

app.add_url_rule(
    '/ajax/reconcile/<int:reconcile_id>',
    view_func=TxnReconcileAjax.as_view('txn_reconcile_ajax')
)

app.add_url_rule(
    '/ajax/unreconciled/ofx',
    view_func=OfxUnreconciledAjax.as_view('ofx_unreconciled_ajax')
)

app.add_url_rule(
    '/ajax/unreconciled/trans',
    view_func=TransUnreconciledAjax.as_view('trans_unreconciled_ajax')
)

app.add_url_rule(
    '/ajax/reconcile',
    view_func=ReconcileAjax.as_view('reconcile_ajax')
)
